/*
    GnuGoS60 - A Port of GNU Go to Series60/Symbian
    Copyright (C) 2006, 2007  Ian Dunbar

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include <eikenv.h>
#include <gnugo.rsg>
#ifdef UIQ3
#include <gnugo.mbg>
#include <qikappui.h>
#include <devicekeys.h>
#else
#include <aknutils.h> 
#include <aknviewappui.h> 
#include <aknappui.h> 
#include <aknnotewrappers.h> 
#endif

#include "GoGame.h"
#include "Debug.h"
#include "BoardControl.h"
#include "GnuGoAppView.h"

// Standard construction sequence
CBoardControl* CBoardControl::NewL(CgnugoAppView &aContainer, const TRect& aRect, class GoGame *aGame)
{
    DEBUG_PRINTF (("CBoardControl::NewL"));
	CBoardControl* self = CBoardControl::NewLC(aContainer, aRect, aGame);
	CleanupStack::Pop(self);
	return self;
}

CBoardControl* CBoardControl::NewLC(CgnugoAppView &aContainer, const TRect& aRect, class GoGame *aGame)
{
    DEBUG_PRINTF (("CBoardControl::NewLC"));
	CBoardControl* self = new (ELeave) CBoardControl;
	CleanupStack::PushL(self);
	self->ConstructL(aContainer, aRect, aGame);
	return self;
}

CBoardControl::CBoardControl()
{
	// no implementation required
}

CBoardControl::~CBoardControl()
{
#ifdef UIQ3
delete iBoardBitmap;
for(TInt loop = 0;loop<6;loop++)
{
	delete iWhiteNodes[loop];
	delete iBlackNodes[loop];
}
#endif
}

void CBoardControl::NewGame (int boardSize)
{
    DEBUG_PRINTF (("CBoardControl::NewGame"));
    iGame->NewGame (boardSize);
    SizeChanged ();
    DrawNow ();
}

void CBoardControl::ConstructL(CgnugoAppView &aContainer, const TRect& aRect, class GoGame *aGame)
{
    DEBUG_PRINTF (("CBoardControl::ConstructL"));
    iGame = aGame;
	iMainWindow = &aContainer;
#ifdef UIQ3
	iBoardBitmap = iEikonEnv->CreateBitmapL(_L("*"), EMbmGnugoBoard);
	for(TInt loop = 0;loop<6;loop++)
	{
		iWhiteNodes[loop] =  iEikonEnv->CreateBitmapL(_L("*"), EMbmGnugoNode_white_11x11+loop);
		iBlackNodes[loop] = iEikonEnv->CreateBitmapL(_L("*"), EMbmGnugoNode_black_11x11+loop);
	}
	
#endif

    TUint size = iGame->BoardSize ();
	// set the current position of the cursor
	iCurrentX = iCurrentY = (size+1)/2;

	// sets the container window
	SetContainerWindowL ((const CCoeControl&) aContainer);

	// Set the windows size
	SetRect(aRect);

	// Set the default player
	SetPlayer (Player::White);

	// Activate the window, which makes it ready to be drawn
	ActivateL();
}

TPoint CBoardControl::Vertex (TUint x, TUint y) const
{
	// Vertex 0,0 is the top left corner of the board
	// Vertex 20,20 is the bottom right corner of the board
	// vertex x,y are vertices 1..19 of the board
        TUint size = iGame->BoardSize ();
	
	// limit the extents to the maximum
	if (x > size+1) x = size+1;
	if (y > size+1) y = size+1;

	// calculate the point coordinates of the vertex
	TPoint point (centeringX + cellWidth * x, centeringY + cellHeight * y);

	return point;
}

// Draw this application's view to the screen
void CBoardControl::Draw(const TRect& /* aRect */) const
{
    // Get the standard graphics context 
    CWindowGc& gc = SystemGc();

    // make sure we don't draw outside of control's own area
    gc.SetClippingRect (Rect());

    // put the origin at the top left of the control's area
    gc.SetOrigin (Rect().iTl);

    TUint size = iGame->BoardSize ();

    // Draw the yellow board centered in the area of the control
    gc.SetBrushStyle (CGraphicsContext::ESolidBrush);
    gc.SetBrushColor (TRgb (248, 191, 36));
    gc.SetPenStyle (CGraphicsContext::ENullPen);
#ifdef UIQ3
	gc.BitBlt(Vertex(0,0), iBoardBitmap);
#else
    gc.DrawRect (TRect (Vertex (0,0), Vertex (size+1,size+1) + TPoint (1,1)));
#endif
    // Draw the black surround 1 pixel outside the edge vertices
    gc.SetPenStyle (CGraphicsContext::ESolidPen);
    gc.SetBrushStyle (CGraphicsContext::ENullBrush);
    gc.DrawRect (TRect (Vertex (1,1) - TPoint (1,1), Vertex (size,size) + TPoint (2,2)));

    // draw the 19x19 grid of lines
    {
        for (int i = 1; i <= size; ++i) {
            gc.DrawLine (Vertex (i, 1), Vertex (i, size) + TPoint (0,1));
            gc.DrawLine (Vertex (1, i), Vertex (size, i) + TPoint (1,0));
        }
    }

    {
        for (int x = 1; x <=size; x++) {
            for (int y = 1; y <=size; y++) {
                DrawDot (x, y);
            }
        }
    }

    {
        for (int x = 1; x <=size; ++x) {
            for (int y = 1; y <=size; ++y) {
                DrawCounter (x, y);
            }
        }
    }

	TPoint computerMove = iGame->LastComputerMove();
	if(computerMove.iX != -1 && computerMove.iY != 1)
	{
		DrawCursor (computerMove.iX+1, computerMove.iY+1, ETrue);
	}

    DrawCursor (iCurrentX, iCurrentY);
}

void CBoardControl::DrawCursor (TUint x, TUint y,TBool aComputerCursor) const
{
	// Get the standard graphics context 
	CWindowGc& gc = SystemGc();

	TPoint tL = Vertex (x,y) - TPoint (2 * cellWidth / 3, 2 * cellHeight / 3);
	TPoint bR = Vertex (x,y) + TPoint (2 * cellWidth / 3, 2 * cellHeight / 3);

	gc.SetPenStyle (CGraphicsContext::ESolidPen);
	if(aComputerCursor)
	{
	gc.SetPenColor (TRgb (255,0,0));
	}
	else
	{
	gc.SetPenColor (TRgb (0,255,0));
	}
	gc.SetBrushStyle (CGraphicsContext::ENullBrush);

	gc.DrawRect (TRect (tL, bR + TPoint (1,1)));
}

void CBoardControl::DrawDot (TUint x, TUint y) const
{
    if (!iGame->IsHandicapSpot (x,y)) return;

    // Get the standard graphics context 
    CWindowGc& gc = SystemGc();

    TUint size = iGame->BoardSize ();

    // trim to board extents
    if (x < 1) x = 1;
    if (y < 1) y = 1;
    if (x > size) x = size;
    if (x > size) x = size;

    gc.SetBrushColor (TRgb (0, 0, 0));
    gc.SetPenColor (TRgb (0, 0, 0));
    gc.SetBrushStyle (CGraphicsContext::ESolidBrush);
    gc.SetPenStyle (CGraphicsContext::ESolidPen);

    TUint height = cellHeight / 4;
    if (1 >= height) height = 2;

    TUint width = cellWidth / 4;
    if (1 >= width) width = 2;

    gc.DrawEllipse (TRect (Vertex (x,y) - TPoint (width, height),
                Vertex (x,y) + TPoint (width + 1, height + 1)));
}

void CBoardControl::SetCounter (Player::Type /*player*/, TUint x, TUint y)
{
    DEBUG_PRINTF (("CBoardControl::SetCounter"));
    TUint size = iGame->BoardSize ();
	if ((x < 1) || (x > size)) return;
	if ((y < 1) || (y > size)) return;
    if (Counter (x,y) != Player::None) return;

	if (iGame) 
	{
		if(!iGame->DoMove (x,y))
		{
			iEikonEnv->InfoMsg(_L("Illegal move"));
		}
	}
}

void CBoardControl::DrawCounter (TUint x, TUint y) const
{
	// Get the standard graphics context 
	CWindowGc& gc = SystemGc();

        TUint size = iGame->BoardSize ();
	// trim to board extents
	if (x < 1) x = 1;
	if (y < 1) y = 1;
	if (x > size) x = size;
	if (x > size) x = size;

	Player::Type player = Counter (x,y);

	if (player == Player::None) return;

	gc.SetPenColor (TRgb (0, 0, 0));
	gc.SetPenStyle (CGraphicsContext::ESolidPen);
#ifdef UIQ3
	gc.SetBrushStyle (CGraphicsContext::ENullBrush);

	TInt index = 0;
	CFbsBitmap* bitmap = NULL;
	CFbsBitmap* mask = NULL;
	if(cellWidth>18)
	{
		index = 2;
	}
	else if(cellWidth > 13)
	{
		index = 1;
	}

	switch (player) {
		case Player::Black:
			bitmap = iBlackNodes[index];
			mask = iBlackNodes[index+3];
			break;
		case Player::White:
			bitmap = iWhiteNodes[index];
			mask = iWhiteNodes[index+3];
			break;
	}
	
	if(bitmap != NULL)
	{
		TSize sz = bitmap->SizeInPixels();
		gc.BitBltMasked(TPoint(Vertex (x,y) - TPoint (sz.iWidth/2, sz.iHeight/2)), bitmap, TRect(TPoint(0,0),sz), mask, EFalse);
	}
#else
	gc.SetBrushStyle (CGraphicsContext::ESolidBrush);

	switch (player) {
		case Player::Black:
			gc.SetBrushColor (TRgb (0, 0, 0));
			break;
		case Player::White:
			gc.SetBrushColor (TRgb (255, 255, 255));
			break;
		default:
			break;
	}

	gc.DrawEllipse (TRect (Vertex (x,y) - TPoint (cellWidth/2, cellHeight/2),
			       Vertex (x,y) + TPoint (cellWidth/2 + 1, cellHeight/2 + 1)));
#endif
}

void CBoardControl::SizeChanged ()
{
    DEBUG_PRINTF (("CBoardControl::SizeChanged"));
	// get the new size;
	TRect rect = Rect ();
    TUint size = iGame->BoardSize ();

	iBoardWidth = rect.Width ();
	cellWidth = iBoardWidth / (size+1);
	centeringX = (iBoardWidth % (size+1)) / 2;

	iBoardHeight = rect.Height ();
	cellHeight = iBoardHeight / (size+1);
	centeringY = (iBoardHeight % (size+1)) / 2;

    iCurrentX = (size +1) / 2;
    iCurrentY = (size +1) / 2;
}

TCoeInputCapabilities CBoardControl::InputCapabilities() const
{
    DEBUG_PRINTF (("CBoardControl::InputCapabilities"));
	TCoeInputCapabilities capabilities (TCoeInputCapabilities::ENavigation);

	return capabilities;
}

void CBoardControl::SetCursor (TUint x, TUint y)
{
    DEBUG_PRINTF (("CBoardControl::SetCursor"));
    TUint size = iGame->BoardSize ();
	iCurrentX = x;
	if (iCurrentX > size) iCurrentX = 1;
	if (iCurrentX < 1) iCurrentX = size;

	iCurrentY = y;
	if (iCurrentY > size) iCurrentY = 1;
	if (iCurrentY < 1) iCurrentY = size;
}

void CBoardControl::SetPlayer (Player::Type player)
{
	currentPlayer = player;
}

CBoardControl::Player::Type CBoardControl::CyclePlayer (Player::Type player) const
{
	player = (Player::Type) (((TUint) player) + 1);
	if (player >= Player::Count) {
		player = Player::None;
	}
	return player;
}

CBoardControl::Player::Type CBoardControl::Counter (TUint x, TUint y) const
{
        TUint size = iGame->BoardSize ();
	if ((x < 1) || (x > size)) return Player::None;
	if ((y < 1)|| (y > size)) return Player::None;

	if (iGame == 0) return Player::None;

	GoGame::PlayerType type = iGame->Vertex(x,y);
	switch (type) {
		case GoGame::Black:
			return Player::Black;
		case GoGame::White:
			return Player::White;
		default:
			return Player::None;
	}
}

/**
 *  From CCoeControl
 *  Handles pointer events
 */	  
void CBoardControl::HandlePointerEventL(const TPointerEvent& aPointerEvent)
{
	if(aPointerEvent.iType == TPointerEvent::EButton1Down)
	{
		if(!iGame->ComputerThinking())
		{
			TPoint pos = (aPointerEvent.iPosition - Position())-TPoint(cellWidth/2, cellHeight/2);
			TInt newCurrentX=1+ (pos.iX/cellWidth);
			TInt newCurrentY =1+ (pos.iY/cellHeight);

			if(newCurrentX <= iGame->BoardSize () && newCurrentY <= iGame->BoardSize () && iGame->GameState() != GoGame::ELoadedGameState)
			{
				if((newCurrentX != iCurrentX || newCurrentY != iCurrentY))
				{
					iCurrentX = newCurrentX;
					iCurrentY = newCurrentY;
					SetCursor (iCurrentX, iCurrentY);
				}
				else
				{
					SetCounter (CyclePlayer (Counter (iCurrentX, iCurrentY)), iCurrentX, iCurrentY);
				}
				DrawNow();
			}
		}
	}
}


TKeyResponse CBoardControl::OfferKeyEventL (const TKeyEvent &aKeyEvent, TEventCode aType)
{
    DEBUG_PRINTF (("CBoardControl::OfferKeyEventL"));
	TKeyResponse response = EKeyWasNotConsumed;
	if (aType == EEventKey && !iGame->ComputerThinking()) {
		switch (aKeyEvent.iScanCode) {
#ifdef UIQ3
			
			case EStdDeviceKeyFourWayLeft:
			case EStdKeyNkp4:
			case '4':
			case 'D':
			case 'F':
#endif
			case EStdKeyLeftArrow:
				response = EKeyWasConsumed;
				if(iGame->GameState() == GoGame::ELoadedGameState)
				{
					iGame->PrevMove();
				}
				else
				{

					SetCursor (iCurrentX - 1, iCurrentY);
				}
				break;
#ifdef UIQ3
			case EStdDeviceKeyFourWayRight:
			case EStdKeyNkp6:
			case '6':
			case 'J':
			case 'K':
#endif
			case EStdKeyRightArrow:
				response = EKeyWasConsumed;
				
				if(iGame->GameState() == GoGame::ELoadedGameState)
				{
					iGame->NextMove();
				}
				else
				{
					SetCursor (iCurrentX + 1, iCurrentY);
				}
				break;
#ifdef UIQ3
			case EStdDeviceKeyFourWayUp:
			case EStdDeviceKeyTwoWayUp:
			case EStdKeyNkp2:
			case '2':
			case 'T':
			case 'Y':
#endif
			case EStdKeyUpArrow:
				response = EKeyWasConsumed;
				if(iGame->GameState() == GoGame::ELoadedGameState)
				{
					iGame->PrevMove();
				}
				else
				{
					SetCursor (iCurrentX, iCurrentY - 1);
				}
				break;
#ifdef UIQ3
			case EStdDeviceKeyFourWayDown:
			case EStdDeviceKeyTwoWayDown:
			case EStdKeyNkp8:
			case '8':
			case 'V':
			case 'B':
#endif	
			case EStdKeyDownArrow:
				response = EKeyWasConsumed;
				if(iGame->GameState() == GoGame::ELoadedGameState)
				{
					iGame->NextMove();
				}
				else
				{
					SetCursor (iCurrentX, iCurrentY + 1);
				}
				break;
#ifdef UIQ3
			case EStdDeviceKeySoftkey1:
			case EStdDeviceKeyAction:
			case EStdDeviceKeyDone:
			case EStdKeyNkp5:
			case '5':
			case 'G':
			case 'H':
#endif	
			case EStdKeySpace:
			case EStdKeyDevice3:
				if(iGame->GameState() != GoGame::ELoadedGameState)
				{
					response = EKeyWasConsumed;
					SetCounter (CyclePlayer (Counter (iCurrentX, iCurrentY)), iCurrentX, iCurrentY);
				}
				break;
#ifdef UIQ3
			case EStdDeviceKeyClear:
#endif
			case EStdKeyBackspace:
				if(iGame->GameState() != GoGame::ELoadedGameState)
				{
					response = EKeyWasConsumed;
					SetCounter (Player::None, iCurrentX, iCurrentY);
				}
				break;
			default:
#if 0
				// Print our the current key press
				_LIT(message,"Key %d");
				HBufC * msg = HBufC::NewL (100);
				TPtr ptr (0,0);
				ptr.Set (msg->Des ());
				ptr.Format (message, aKeyEvent.iScanCode);
				CAknInformationNote* informationNote = new (ELeave) CAknInformationNote;
				informationNote->ExecuteLD(ptr);
#endif

				break;
		}
		if (EKeyWasConsumed == response) DrawNow ();
	}
	return response;
}

